# -*- coding: utf-8 -*-
import datetime
import logging
import json
import uuid
import werkzeug

from odoo import api, models, fields

from odoo.addons.saas_worker.compat import request
from odoo.exceptions import UserError
from odoo.tools.translate import _

_logger = logging.getLogger(__name__)

class AuthOauth2Token(models.TransientModel):
    """Oauth2 Token Class"""

    _name = 'auth.oauth2.token'
    _description = 'OAuth2 Token'

    TOKEN_EXPIRATION_TIMESPAN = 3600  # in seconds
    DEFAULT_TOKEN_SCOPES = ['userinfo']
    ENTERPRISE_SCOPES = ['apps', 'migration', 'support', 'trial']

    def _default_expiration(self):
        return fields.Datetime.to_string(self._expiration_in(self.TOKEN_EXPIRATION_TIMESPAN))

    def _expiration_in(self, seconds):
        return datetime.datetime.utcnow() + datetime.timedelta(seconds=seconds)

    token = fields.Char('Token', size=32, readonly=True, index=True)
    user_id = fields.Many2one('res.users', 'User', index=True, readonly=True)
    client = fields.Char('Client', help="Client Application for which the token has been generated", readonly=True, index=True)  # TODO: auth.oauth2.client object
    scope = fields.Char('Scope', help="Scope for which the token has ben delivered", readonly=True, index=True)
    expires_at = fields.Datetime('Token expiration date', default=_default_expiration, index=True)
    error_message = fields.Char('Error message if token is invalid')
    skip_payment = fields.Boolean('Payment Skipped')
    support = fields.Boolean('Support token')
    support_user_id = fields.Many2one('res.users', 'Support User', readonly=True)

    def _get_scope_method(self, scope):
        key = '_scope_%s' % scope
        return getattr(self, key, None)

    def _has_scope(self, scope):
        m = self._get_scope_method(scope)
        return m is not None

    def _sanitize_scope(self, scope=""):
        scopes = scope.split()
        scopes = filter(self._has_scope, scopes)    # keep the ones that are valid
        if not scopes:
            scopes = self.DEFAULT_TOKEN_SCOPES
        return ' '.join(scopes)

    def _is_enterprise_scope(self, scope=""):
        scopes = scope.split()
        return any(s in self.ENTERPRISE_SCOPES for s in scopes)

    def _get_buy_wizard_action(self, client_id, scope, params):
        redirect = '/oauth2/auth/?' + werkzeug.urls.url_encode(dict(params))
        return {'redirect_uri': '/odoo-enterprise/upgrade-online/%s?%s' % (client_id, werkzeug.urls.url_encode({'redirect': redirect}))}

    @api.model
    def get_scope_manifests(self, scope=""):
        scopes = scope.split()
        if not scopes:
            scopes = self.DEFAULT_TOKEN_SCOPES
        manifests = []
        for s in scopes:
            m = self._get_scope_method(s)
            if not m:
                raise UserError(_("This OAuth provider does not support scope '%s'.", s))
            manifests.append(m.__doc__)
        return manifests

    def _register_db(self, client_id="", dbname="", url="", mode="", register_user=False):
        Databases = self.env['openerp.enterprise.database']
        db_id = Databases.sudo().search([('db_uuid', '=', client_id)], limit=1)
        if db_id:
            values = {}
            if dbname:
                values['db_name'] = dbname
            if url:
                values['url'] = url
            if mode:
                values['mode'] = mode
            if values:
                db_id.write(values)
        else:
            data = {
                'db_uuid': client_id,
                'first_ping': fields.Date.context_today(self.env.user),
                'last_ping': fields.Date.context_today(self.env.user),
                'db_name': dbname,
                'url': url,
                'mode': mode,
            }
            db_id = Databases.sudo().create(data)
        if register_user and register_user not in db_id.registered_user_ids.mapped('user_id').ids:
            db_id.write({'registered_user_ids': [(0, 0, {'user_id': register_user})]})

        return db_id

    def _need_to_pay(self, db, scope):
        if not self._is_enterprise_scope(scope):
            return False

        if 'apps' in scope.split():
            return False

        if db.subscription_id.subsciption_state in ('3_progress', '4_paused'):
            return False

        return True

    @api.model
    def get_token(self, params):
        """Return a new OAUth2 token for the requested scope, registering the client on the
        fly if needed.

        This does NOT verify that the token can be used for any specific kind of
        operation, it is merely used to prove the identity of the requesting user
        to an external service, even if the `client_id` is a proxy for the database UUID.
        The end-service is still responsible for verifying that the user is authorized
        to perform the final operation on the target database.
        This is typically done by calling `database._check_and_log_spoof()` or remotely via
        the `/odoo-enterprise/staff-access` route.
        """
        client_id = params.get('client_id')
        scope = self._sanitize_scope(params.get('scope'))

        if not client_id:
            return self.sudo().get_invalid_token(client_id, scope, 'Missing client_id')
        dbname = ''
        url = ''
        try:
            state = json.loads(params.get('state'))
            dbname = ("d" in state) and state['d'] or ''
            url = ("u" in state) and state['u'] or ''
        except Exception:
            pass

        if 'trial' in scope:
            params['skip_payment'] = True
            mode = 'saas trial15days'
        elif 'apps' in scope:
            mode = 'onsite has_apps'
        else:
            mode = None

        db = self._register_db(client_id=client_id, dbname=dbname, url=url, mode=mode, register_user=False)

        if self._need_to_pay(db, scope) and not params.get('skip_payment'):
            return self._get_buy_wizard_action(client_id, scope, params)

        return self._create_token(client_id, self.env.uid, scope, None, params.get('skip_payment'))

    @api.model
    def _get_trial_token(self, user_id, db_uuid, session_info=None):
        """
        Creates an authentication token to identify the current user against a trial database
        @arg user_id          integer  Id of the user associated to the token
        @arg db_uuid          string   UUID of the database that shall verify this authentication
        @arg session_info     dict     Information normally found in request.session, used where there is no http.request.
                                       If provided, key/values should be:
                                       - impostor  request.session.impostor
                                       - login     request.session.login
                                       - uid       request.session.uid
        """
        token_info = self._create_token(db_uuid, user_id, 'trial', False, False, validity_seconds=60, session_info=session_info)
        return token_info['access_token']

    @api.model
    def get_invalid_token(self, client_id, scope, error_message):
        return self._create_token(client_id, False, scope, error_message, True)

    def _create_token(self, client_id, user_id, scope, error_message, skip_payment, validity_seconds=None, session_info=None):
        """
        Creates an authentication token associated with the given information
        @arg client_id        string   Client Application for which the token has been generated
        @arg user_id          integer  Id of the user associated to the token
        @arg scope            string   Scope for which the token has ben delivered
        @arg error_message    string   Error message if token is invalid
        @arg skip_payment     boolean  Payment Skipped
        @arg validity_seconds float    How many seconds this token will be considered as valid.
                                       Defaults to 1 hour.
        @arg session_info     dict     Information normally found in request.session, used where there is no http.request.
                                       If provided, key/values should be:
                                       - impostor  request.session.impostor
                                       - login     request.session.login
                                       - uid       request.session.uid
        """
        scope = self._sanitize_scope(scope)
        token = str(uuid.uuid4()).replace('-', '')

        if validity_seconds is None:
            validity_seconds = self.TOKEN_EXPIRATION_TIMESPAN

        values = {
            'token': token,
            'user_id': user_id,
            'client': client_id,
            'scope': scope,
            'skip_payment': skip_payment,
            'expires_at': self._expiration_in(validity_seconds)
        }
        if error_message:
            values['error_message'] = error_message

        if session_info is None:
            if not request:
                raise Exception('No http.request and no session_info argument')
            session_info = request.session

        impostor = session_info.get('impostor')
        if impostor:
            values['support_user_id'] = impostor
            if self.env['res.users'].browse(impostor).has_group('openerp_enterprise.group_manager'):
                values['support'] = True
        if 'sale' in scope:
            if self.env.user.has_group('sales_team.group_sale_salesman'):
                values['support'] = True

        if self.env.su:
            self.create(values)
        else:
            # sudo mode: only allow self-creation of tokens or invalid tokens (user_id=False)
            assert values["user_id"] in (False, self.env.uid)
            self.sudo().create(values)

        log_params = {
            "scope": scope,
            "login": session_info.get('login'),
            "uid": session_info.get('uid'),
            "impostor": session_info.get('impostor'),
            "ip": request.httprequest.environ['REMOTE_ADDR'] if request else False,
            "exp": values["expires_at"],
            "db_uuid": client_id,
        }
        _logger.info("New token: %(scope)s | %(login)s@%(db_uuid)s | uid #%(uid)s | ip %(ip)s | exp %(exp)s", log_params)

        return {
            'access_token': token,
            'scope': scope,
            'expires_in': validity_seconds,
        }

    @api.model
    def tokeninfo(self, access_token):
        token = self.search([('token', '=', access_token)], limit=1)
        if not token:
            return {"error": "invalid_token"}

        expires_in = int(round((token.expires_at - datetime.datetime.utcnow()).total_seconds()))

        if expires_in <= 0:
            return {"error": "invalid_token"}

        res = {
            "user_id": token.user_id.id,
            "scope": token.scope,
            "expires_in": expires_in,
            "audience": token.client,
            "support": token.support,
        }
        if token.error_message:
            res['error'] = token.error_message
        if token.skip_payment:
            res['skip_payment'] = True

        mode = None
        if 'trial' in token.scope and not res['user_id']:
            mode = 'saas trial'

        # Do *not* register the user at this point, as we don't know if the UID is granted access
        # on the OAuth client side. User will be registered via ping later if needed.
        # This database registration/creation flow is deprecated, and this whole part of the code
        # should be reviewed and cleaned up!
        db = self._register_db(token.client, mode=mode, register_user=False)

        for s in token.scope.split():
            m = self._get_scope_method(s)
            m(res, token, db)
        return res

    def _scope_userinfo(self, res, token, db):
        """Basic information about your account"""
        for k in ['email', 'name']:
            v = token.user_id[k]
            if v:
                res[k] = v

    def _scope_apps(self, res, token, db):
        # Everybody have access to apps now
        res['apps'] = True

    def _scope_migration(self, res, token, db):
        res['migration'] = db.subscription_id and db.subscription_id.enterprise_is_enterprise

    def _scope_support(self, res, token, db):
        res['support'] = token.support or db.subscription_id and db.subscription_id.enterprise_is_enterprise

    def _scope_trial(self, res, token, db):
        res['trial'] = True

    def _scope_skiptheuse(self, res, token, db):
        res['skiptheuse'] = True

    def _scope_sale(self, res, token, db):
        res['sale'] = True

    @api.autovacuum
    def _garbage_collector(self):
        """Unlink expired tokens."""
        # Do it in raw SQL as deletion is now too slow using unlink()
        # and we know there are no auxiliary tables to cascade for
        # this pure system table.
        self._cr.execute(f"""
            DELETE FROM {self._table} WHERE expires_at < NOW() AT TIME ZONE 'UTC'
        """)
        _logger.info("Deleted %s expired oauth2 tokens", self._cr.rowcount)
